"""All classes in this file provide similar signatures.

They all have :code:`text: str` and
:code:`style: Callable[[Union[str, Element]], Element])` arguments.

:code:`style` is a factory that will be applied to the element.
"""

from typing import Callable, Union

from .bases import Element, PlainText, Text


class Link(Element):
    """Inline link element.

    Args:
        text (str):
            Text of link.
        url (str):
            Web address on the internet.
        style (Callable[[Union[str, Element]], Element]):
            Style factory which will be applied to the element.
            :class:`telegram_text.bases.PlainText` by default.
    """

    def __init__(self, text: str, url: str, style: Callable[[Union[str, Element]], Element] = PlainText):
        self.text: Element = style(text)
        self.url = url

    def to_plain_text(self) -> str:
        return f"{self.text.to_plain_text()} ({self.url})"

    def to_markdown(self) -> str:
        return f"[{self.text.to_markdown()}]({self.url})"

    def to_html(self) -> str:
        return f'<a href="{self.url}">{self.text.to_html()}</a>'


class InlineUser(Link):
    """Inline link to user with specific for Telegram url
    (:code:`tg://user?id={}`).

    Args:
        text (str):
            Text of link.
        user_id (int):
            Identifier of user.
        style (Callable[[Union[str, Element]], Element]):
            Style factory which will be applied to the element.
            :class:`telegram_text.bases.PlainText` by default.
    """

    def __init__(self, text: str, user_id: int, style: Callable[[Union[str, Element]], Element] = Text):
        url = f"tg://user?id={user_id}"
        super().__init__(text=text, url=url, style=style)


class _Reference(Element):
    def __init__(self, text: str, style: Callable[[str], Element] = PlainText):
        self.text = style(text)

    def to_plain_text(self) -> str:
        return self.text.to_plain_text()

    def to_markdown(self) -> str:
        return self.text.to_markdown()

    def to_html(self) -> str:
        return self.text.to_html()


class User(_Reference):
    """Link to user by nickname.

    Args:
        text (str):
            Nickname of user.
        style (Callable[[Union[str, Element]], Element]):
            Style factory which will be applied to the element.
            :class:`telegram_text.bases.PlainText` by default.
    """

    def __init__(self, text: str, style: Callable[[Union[str, Element]], Element] = PlainText):
        text = '@' + text.lstrip('@')
        super().__init__(text, style=style)


class Hashtag(_Reference):
    """Hashtag link element.

    Args:
        text (str):
            Hashtag name.
        style (Callable[[Union[str, Element]], Element]):
            Style factory which will be applied to the element.
            :class:`telegram_text.bases.PlainText` by default.
    """

    def __init__(self, text: str, style: Callable[[Union[str, Element]], Element] = PlainText):
        text = '#' + text.lstrip('#')
        super().__init__(text, style=style)


class Emoji(Element):
    """Custom Telegram emoji.

    Custom emoji entities can only be used by bots that purchased additional
    usernames on `Fragment <https://fragment.com/>`_.

    Args:
        emoji_id (int):
            Id of the custom emoji.
        default (str):
            Alternative value for the custom emoji.
            The emoji will be shown instead of the custom emoji in places
            where a custom emoji cannot be displayed (e.g., system
            notifications) or if the message is forwarded by a non-premium user.
    """

    def __init__(self, emoji_id: int, default: str):
        self.emoji_id = emoji_id
        self.default = default

    def to_plain_text(self) -> str:
        return self.default

    def to_markdown(self) -> str:
        return f"![{self.default}](tg://emoji?id={self.emoji_id})"

    def to_html(self) -> str:
        return f'<tg-emoji emoji-id="{self.emoji_id}">{self.default}</tg-emoji>'



"""Telegram markdown and official markdown standards are very different.
Telegram markdown has new elements and hasn't some main old elements from
the official markdown format. This file is a compromise that adds missing
elements to Telegram.
"""

from typing import Callable

from .bases import Chain, Element, NEW_LINE
from .styles import Bold, PlainText


class UnorderedList(Chain):
    """Unordered (bullet) list.

    Example:

    * First element.
    * Second element.
    * Third element.

    Args:
        elements (Tuple[Element]):
            Elements for the list. Every element is a single point.
        style (Callable[[str], Element]):
            Style factory that will be applied to every point
            (only special characters of list, not passed elements).
            :class:`telegram_text.styles.Bold` by default.
        sep (str):
            Separator between every point.
            :const:`telegram_text.bases.NEW_LINE` by default.
    """

    def __init__(self, *elements: Element, style: Callable[[str], Element] = Bold, sep: str = NEW_LINE):
        points = [
            style("*") + element
            for element in elements
        ]
        super().__init__(*points, sep=sep)


class OrderedList(Chain):
    """Ordered (numbered) list.

    Example:

    1. First element.
    2. Second element.
    3. Third element.

    Args:
        elements (Tuple[Element]):
            Elements for the list. Every element is a single point.
        style (Callable[[str], Element]):
            Style factory that will be applied to every point
            (only special characters of list, not passed elements).
            :class:`telegram_text.bases.PlainText` by default.
        sep (str):
            Separator between every point.
            :const:`telegram_text.bases.NEW_LINE` by default.
    """

    def __init__(self, *elements: Element, style: Callable[[str], Element] = PlainText, sep: str = NEW_LINE):
        points = [
            style(f"{n}.") + element
            for n, element in enumerate(elements, start=1)
        ]
        super().__init__(*points, sep=sep)



from typing import Union

from abc import ABC, abstractmethod

NEW_LINE = '\n'  #: New line character constant.
SPACE = ' '  #: Space character constant.


class AbstractElement(ABC):
    """The interface every component implements."""

    @abstractmethod
    def __add__(self, other: Union[str, "Element"]) -> "Chain":
        """Return a new :class:`Chain` object when sum
        :code:`<Element object> + <Element object>`. You can also use it with
        :obj:`str` object, then this method will wrap a :obj:`str` with
        :class:`PlainText` object.
        """
        raise NotImplementedError

    @abstractmethod
    def to_plain_text(self) -> str:
        """Format the element to plain text without escaping, tags or special
        characters.
        """
        raise NotImplementedError

    @abstractmethod
    def to_markdown(self) -> str:
        """Format the element to Markdown/MarkdownV2 format according to
        Telegram specification with escaping if necessary.
        """
        raise NotImplementedError

    @abstractmethod
    def to_html(self) -> str:
        """Format the element to html according to Telegram specification."""
        raise NotImplementedError


class Element(AbstractElement, ABC):
    """Base class every component must inherit."""

    def __add__(self, other: Union[str, "Element"]) -> "Chain":
        if isinstance(other, str):
            other = PlainText(other)
        return Chain(self, other)

    def __radd__(self, other: str) -> "Chain":
        return PlainText(other).__add__(self)

    def __eq__(self, other: object) -> bool:
        """Equality function to write
        :code:`<Element object> == <Element object>`.
        """
        if not isinstance(other, Element):
            return NotImplemented
        return type(self) is type(other) and self.to_plain_text() == other.to_plain_text()

    def __str__(self) -> str:
        """Call :meth:`.to_markdown` function."""
        return self.to_markdown()


class Text(Element):
    """Basic text component. We don't recommend you use this outside the code
    of this module because this class doesn't include escaping logic, it means
    Telegram API may reject you.
    """

    def __init__(self, text: Union[str, Element]):
        if isinstance(text, Element):
            text = text.to_plain_text()
        self.text: str = text

    def to_plain_text(self) -> str:
        return self.text

    def to_markdown(self) -> str:
        """Return original text without escaping."""
        return self.text

    def to_html(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: '{self.text}'>"


class PlainText(Text):
    """Basic text element you can safely use. This element implements escaping
    logic.
    """

    escaping_chars = ('_', '*', '[', ']', '(', ')', '~', '`', '>', '#', '+', '-', '=', '|', '{', '}', '.', '!')
    """Tuple of character that will be escaped according to Telegram specification."""

    def _escape(self, text: str) -> str:
        escaping_prefix = '\\'
        mapping = str.maketrans({char: escaping_prefix + char for char in self.escaping_chars})
        return "".join(char.translate(mapping) for char in text)

    def to_markdown(self) -> str:
        """Format the element to Markdown/MarkdownV2 format according to
        Telegram specification with escaping if necessary.
        """
        return self._escape(self.text)


class Chain(Element):
    """Combination between a few elements.

    Args:
        elements (Tuple[Element]): Any objects with :class:`AbstractElement` interface.
        sep (str): The separator between elements. :const:`SPACE` by default.
    """

    def __init__(self, *elements: Element, sep: str = SPACE):
        self.elements = elements
        self.sep = sep

    def __add__(self, other: Union[str, Element]) -> "Chain":
        if isinstance(other, str):
            other = PlainText(other)

        if self.sep == SPACE:  # Optimization for default separator
            return Chain(*self.elements, other)
        return Chain(self, other)

    def __contains__(self, item: Element) -> bool:
        """Magic method to write :code:`if <Element object> in <Chain object>`."""
        return item in self.elements

    def to_plain_text(self) -> str:
        return self.sep.join(element.to_plain_text() for element in self.elements)

    def to_markdown(self) -> str:
        return self.sep.join(element.to_markdown() for element in self.elements)

    def to_html(self) -> str:
        return self.sep.join(element.to_html() for element in self.elements)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}: "
            f"{' + '.join(repr(element) for element in self.elements)}, "
            f"sep='{self.sep}'>"
        )



from typing import Type, Union

from abc import ABC

from .bases import Element, PlainText, Text


class Style(Element, ABC):
    """Class that inherits each style. it isolates a part of logic how to
    render styles.

    Args:
        text (Union[str, telegram_text.bases.Element]):
            Text or Element to which the style will be applied.
    """

    markdown_symbol: str
    html_tag: str
    html_class: Union[str, None] = None
    base_style_fabric: Union[Type[Text], Type['Style']] = PlainText

    def __init__(self, text: Union[str, Element]):
        if isinstance(text, str):
            text = self.base_style_fabric(text)
        if isinstance(text, Style) and self.__class__ is text.__class__:
            text = text.text
        self.text: Element = text

    def to_plain_text(self) -> str:
        return self.text.to_plain_text()

    def to_markdown(self) -> str:
        return f"{self.markdown_symbol}{self.text.to_markdown()}{self.markdown_symbol}"

    def to_html(self) -> str:
        class_str = f' class="{self.html_class}"' if self.html_class else ''
        return f'<{self.html_tag}{class_str}>{self.text.to_html()}</{self.html_tag}>'

    def __repr__(self) -> str:
        text = f"'{self.text}'" if isinstance(self.text, self.base_style_fabric) else repr(self.text)
        return f"<{self.__class__.__name__}: {text}>"


class Bold(Style):
    """Bold text. Example: **bold text**."""

    markdown_symbol = '*'
    html_tag = 'b'


class Italic(Style):
    """Italic text. Example: *italic text*."""

    markdown_symbol = '_'
    html_tag = 'i'


class Underline(Style):
    """Underline text. Example: :underline:`underline text`."""

    markdown_symbol = '__'
    html_tag = 'u'


class Strikethrough(Style):
    """Strikethrough text. Example: :strike:`strikethrough text`."""

    markdown_symbol = '~'
    html_tag = 's'


class Spoiler(Style):
    """Spoiler text. We can't provide an example because it's a very specific
    for Telegram messenger formatting.
    """

    markdown_symbol = '||'
    html_tag = 'span'
    html_class = 'tg-spoiler'


class InlineCode(Style):
    """Inline code text. Example: :code:`inline code`."""

    markdown_symbol = '`'
    html_tag = 'code'
    base_style_fabric = Text


class Code(Style):
    """Code block for many lines of text. Telegram doesn't support frontend
    language-specific highlights, but according to documentation, it provides
    an opportunity to specify a language. Perhaps, the Telegram team will add
    support for this in the future.

    Args:
        text (str):
            Text of code block.
        language (Optional[str]):
            Leave it empty if you don't want to specify a language.
    """

    markdown_symbol = '```'
    html_tag = 'code'
    base_style_fabric = Text

    def __init__(self, text: str, language: Union[str, None] = None):
        super().__init__(text)
        self.language: str = language or ''
        self.html_class: Union[str, None] = f'language-{language}' if language else None

    def to_markdown(self) -> str:
        return f"{self.markdown_symbol}{self.language}\n{self.text.to_markdown()}{self.markdown_symbol}"

    def to_html(self) -> str:
        # if lang isn't specified, we don't use <code> tag according to Telegram docs
        if self.html_class:
            return f'<pre>{super().to_html()}</pre>'
        return f'<pre>{self.text.to_html()}</pre>'


class Quote(Style):
    """Quote block.
    See an example below.

        Block quotation started

        Block quotation continued

        The last line of the block quotation
    """

    markdown_symbol = '>'
    html_tag = 'blockquote'

    def to_markdown(self) -> str:
        return self.markdown_symbol + f'\n{self.markdown_symbol}'.join(self.text.to_markdown().split('\n'))



"""This file contains unofficial, but very useful markup elements."""

from typing import Callable

from .bases import Chain, Element, NEW_LINE
from .styles import InlineCode


class TOMLSection(Chain):
    """Section element in :code:`.toml` format.

    Args:
        text (str):
            Section name.
        elements (Tuple[Element]):
            Sections elements.
            Separator is :const:`telegram_text.bases.NEW_LINE`.
        style (Callable[[str], Element]):
            Style factory that will be applied to name.
            :class:`telegram_text.styles.InlineCode` by default.
    """

    def __init__(self, text: str, *elements: Element, style: Callable[[str], Element] = InlineCode):
        super().__init__(style(f"[{text}]"), *elements, sep=NEW_LINE)



from .bases import Chain, PlainText, Text
from .custom import TOMLSection
from .elements import Emoji, Hashtag, InlineUser, Link, User
from .markdown import OrderedList, UnorderedList
from .styles import Bold, Code, InlineCode, Italic, Quote, Spoiler, Strikethrough, Underline

__version__ = '0.2.0'
__all__ = [
    "Chain",
    "PlainText",
    "Text",

    "Bold",
    "Code",
    "InlineCode",
    "Italic",
    "Quote",
    "Spoiler",
    "Strikethrough",
    "Underline",

    "Emoji",
    "Hashtag",
    "InlineUser",
    "Link",
    "User",

    "OrderedList",
    "UnorderedList",

    "TOMLSection",
]
